#!/usr/bin/perl

use strict;
use warnings;

package VProtect::Backup;
use Data::Dumper;

use VProtect qw/vim_get_virtual_machines/;
use VMware::Vix::Simple;
use VMware::Vix::API::Constants;

use POSIX "strftime";
use File::Basename;
use LWP::UserAgent;
use LWP::Simple;

my $gos_guestPath;
my $gos_guestVssPgmPath;
my $gos_vmHandle;
my $gos_hostHandle;
my $gos_osLevel;

sub dbg_log {
	my $msg = shift;
	my $time = strftime "%F %H:%M:%S.000 %z", localtime;
	print STDERR $time . ' : D ' . $msg . "\n";
}

sub get_moref_by_uuid ($$$) {
	my $vmuuid = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vmuuid);

	if (!defined $vm_moref) {
		die "VM with UUID " . $vmuuid . " not found.\n" .
		"Try refreshing the list of Virtual Machines\n";
	}

	my $vm_xmlelem = $xmldoc->createElement('vm_info');
	$vm_xmlelem->appendTextChild('vm_moref', $vm_moref->{'value'});
	$vm_xmlelem->appendTextChild('vm_moref_type', $vm_moref->{'type'});
	$xmlroot->addChild($vm_xmlelem);

	return $vm_moref;
}

sub get_template_moref_by_uuid ($$$) {
	my $template_uuid = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $template_moref = VProtect::vim_get_template_by_uuid($template_uuid);

	if (!defined $template_moref) {
		die "Template with UUID " . $template_uuid . " not found.\n" .
		"Try refreshing the list of Virtual Machines\n";
	}

	# Attach similar elements in the XML as those for a VM,
	# so that the complete parsing need not be changed
	my $vm_xmlelem = $xmldoc->createElement('vm_info');
	$vm_xmlelem->appendTextChild('vm_moref', $template_moref->{'value'});
	$vm_xmlelem->appendTextChild('vm_moref_type', $template_moref->{'type'});
	$xmlroot->addChild($vm_xmlelem);

	return $template_moref;
}

sub purge_snaps ($$$);  # required to get type checking
sub purge_snaps ($$$) {
	# Remove snapshots named $snapshot_name
	my ($snapshot_tree, $ch_present_ref, $snapshot_name) = @_;

	if ($snapshot_tree->childSnapshotList) {
		foreach (@{$snapshot_tree->childSnapshotList}) {
			purge_snaps($_, $ch_present_ref, $snapshot_name);
		}
	}

	if ($snapshot_tree->name eq $snapshot_name) {
		my $snap_view = Vim::get_view(mo_ref => $snapshot_tree->snapshot);
		eval {
			$snap_view->RemoveSnapshot(removeChildren => 0);
		}; if ($@) {
			# This is a fatal error.
			my $hdr = "Could not remove previous Unitrends snapshot";

			if (ref($@->detail) eq 'FileLocked') {
				die "$hdr: the file is locked: " . Dumper($@) . "\n";
			}
			elsif (ref($@->detail) eq 'ManagedObjectNotFound') {
				# This should probably not be a fatal error, but we have to be conservative with snapshot removal issues.
				die "$hdr: The object has already been deleted or has not been completely created.\n";
			}
			elsif (ref($@->detail) eq 'NoPermission') {
				die "$hdr: permission denied.  Requires permission " . $@->detail->privilegeId . "\n";
			}
			elsif (ref($@->detail) eq 'SystemError') {
				die "$hdr: " . $@->fault_string . "\n";
			}
			elsif (ref($@->detail) eq 'SystemError') {
				die "Could not remove previous Unitrends snapshot: A general system error occurred: " . $@->detail->reason . "\n";
			}
			else {
				die "$hdr: " . Dumper($@) . "\n";
			}
		}
	}
	elsif ($snapshot_tree->name =~ "Consolidate Helper.*") {
		# Flag that consolidate helper snapshot is present.	
		$$ch_present_ref = 1;
	}
}

# This sub-routine is similar to 'configure_change_tracking' sub-routine
# However, this will be called from 'validate_hw_hash' only for Incr/Diff Backups
sub reset_cbt ($$) {

	my $vm_moref = shift;
	my $xmlroot = shift;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $has_physical_rdm_disks = physical_rdm_disks_present($vm_view);

	if (not $vm_view->capability->changeTrackingSupported) {
		disable_change_tracking($vm_view);
		dbg_log("This VM does not support change tracking.");
		return;
	} else {
		# Change Tracking is supported by the ESX on which this VM is
		if ($vm_view->config->changeTrackingEnabled) { # Safety check
			# Change Tracking is currently enabled on this VM
			dbg_log("  Temporarily disabling change tracking.");
			disable_change_tracking($vm_view);

			dbg_log("  Using snapshot create/delete cycle to stun VM.");
			snapshot_stun_cycle($vm_view);

			dbg_log("  Re-enabling change tracking.");
			enable_change_tracking($vm_view, $xmlroot);
		} else {
			dbg_log("CBT is currently disabled on this VM.");
		}
	}

}

sub validate_hw_hash ($$$$$$$) {

	my $prev_hw_hash = shift;
	my $new_hw_hash = shift;
	my $is_full = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $vm_moref = shift;
	my $hw_hash_using_only_disks = shift;

	if ($is_full) {
		# Don't check
		return;
	}

	# 1.	Condition: $prev_hw_hash ne $new_hw_hash
	# Will run into this condition for the first Backup of a VM after
	# the appliance is upgraded to 9.1
	#
	# 2.	Condition: $prev_hw_hash ne $hw_hash_using_only_disks
	# Will run into this condition from the second Backup for a VM after
	# the appliance is upgraded to 9.1
	#
	# If 1 AND 2 both are NOT met, then, it means that something with the disk
	# configuration of the VM has changed

	if (defined $prev_hw_hash) {
		# Below 2 if statements are only for logging purposes
		if ($prev_hw_hash ne $new_hw_hash) {
			dbg_log("H/W hash passed from previous Backup is not same as the one calculated using old technique.");
		}
		if ($prev_hw_hash ne $hw_hash_using_only_disks) {
			dbg_log("H/W hash passed from previous Backup is not same as the one calculated using new technique.");
		}
	}

	if ( (not defined $prev_hw_hash) or
		( ($prev_hw_hash ne $new_hw_hash) and ($prev_hw_hash ne $hw_hash_using_only_disks) ) ) {

		# Signal that next backup should be promoted to full
		VProtect::promote_next_reason($xmldoc, 'A VM configuration change has been detected.');

		# Reset CBT before dying
		dbg_log("Resetting CBT since a configuration change has been detected.");
		reset_cbt($vm_moref, $xmlroot);

		# Fail backup
		die "VM configuration has changed, a full backup is required.\n";
	}
}

sub check_for_fault_tolerance ($) {
	# Fail the backup if the VM is configured for Fault Tolerance.
	# This is unsupported because snapshots are disabled.

	my $vm_moref = shift;
	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	if (defined $vm_view->config->ftInfo) {
		die "This VM cannot be backed up because it is configured for Fault Tolerance.\n";
	}
}

sub clean_up_snapshots ($$$$) {
	# Remove old 'Unitrends' snapshots left from a previous backup that was terminated uncleanly.
	my ($vm_moref, $xmldoc, $xmlroot, $warning_ref) = @_;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snap_view = $vm_view->snapshot;

	my $ch_present = 0;  # Consolidate helper snapshot is present
	my $snapshot_name = "Unitrends_backup";

	if (defined $vm_view->snapshot) {
		if (@{$vm_view->snapshot->rootSnapshotList}) {
			foreach (@{$vm_view->snapshot->rootSnapshotList}) {
				purge_snaps($_, \$ch_present, $snapshot_name);
			}
		}
	}

	if ($ch_present) {
		my $ch_xmlelem = $xmldoc->createElement('consolidate_helper_info');
		$ch_xmlelem->appendTextChild('consolidate_helper_snapshot_present', 1);
		$xmlroot->addChild($ch_xmlelem);
		$$warning_ref .= "Warning: A consolidate helper snapshot is present on the VM.  Presence of this snapshot indicates a possible issue with the snapshot system for this VM.  Please remove this snapshot to ensure that backups function properly and do not affect the VM.\n\n";
	}
}

sub enable_change_tracking ($$) {
	my $vm_view = shift;
	my $xmlroot = shift;
	
	if ($vm_view->capability->changeTrackingSupported) {
		my $spec = VirtualMachineConfigSpec->new(changeTrackingEnabled => 'true');
		my $task_ref;

		eval {
			# Use a task since ReconfigVM() doesn't wait for completion.
			$task_ref = $vm_view->ReconfigVM_Task(spec => $spec);
		};
		if ($@) {
			if (ref($@) eq 'SoapFault') {
				if (ref($@->detail) eq 'RestrictedVersion') {
					die "Error enabling change tracking on VM: RestrictedVersionFault\n" .
					    "This host is not licensed to perform this operation.\n";
				} elsif (ref($@->detail) eq 'NoPermission') {
					die "Error enabling change tracking on VM: permission to perform this operation was denied.\n" .
					$@->detail->privilegeId . " permission is required\n";
				}
			}

			die "Error enabling change tracking on VM:" . $@ . "\n";
		}

		my $task = Vim::get_view(mo_ref => $task_ref);

		sleep 2;

		while (($task->info->state->val eq 'running') or
				($task->info->state->val eq 'queued')) {
			sleep 2;
			$task->update_view_data();
		}

		if ($task->info->state->val eq 'success') {
			# change tracking was enabled
			$xmlroot->appendTextChild('currentChangeTrackingEnabled', '1');
			dbg_log("Change tracking was enabled");
		} elsif ($task->info->state->val eq 'error') {
			dbg_log("Could not enable change tracking.");
			dbg_log("Dumper of task:\n" . Dumper($task));
			my $error_msg = "<no further information provided>";
			if (defined($task->info->error->localizedMessage)) {
				$error_msg = $task->info->error->localizedMessage;
			}
			die "Could not enable change tracking: $error_msg\n";
		} else {
			die "Unknown state during ReconfigVM task: State->val=" . $task->info->state->val . "\n";
		}
	}
}

sub disable_change_tracking ($) {
	my $vm_view = shift;

	# bug 18622 always turn off changeTrackingEnabled regardless of changeTrackingSuppported.
#	if ($vm_view->capability->changeTrackingSupported) {
		if ($vm_view->config->changeTrackingEnabled) {
			my $spec = VirtualMachineConfigSpec->new(changeTrackingEnabled => 'false');
			# Use a task since ReconfigVM() doesn't wait for completion.
			my $task_ref = $vm_view->ReconfigVM_Task(spec => $spec);
			my $task = Vim::get_view(mo_ref => $task_ref);

			sleep 2;

			while (($task->info->state->val eq 'running') or 
					($task->info->state->val eq 'queued')) {
				sleep 2;
				$task->update_view_data();
			}

			if ($task->info->state->val eq 'success') {
				# change tracking was disabled
				dbg_log("Change tracking was disabled");
			} elsif ($task->info->state->val eq 'error') {
				my $error_msg = "<no further information provided>";
				if (defined($task->info->error->localizedMessage)) {
					$error_msg = $task->info->error->localizedMessage;
				}
				dbg_log("Could not disable change tracking: $error_msg.");
				dbg_log("Dumper of task:\n" . Dumper($task));
				# Do not abort.
			} else {
				die "Disable cbt: State->val=" . $task->info->state->val . "\n";
			}
		}
#	}
}

sub snapshot_stun_cycle ($) {
	# Create a snapshot and remove it to "stun" the VM.
	# Do not quiesce since we aren't using this snapshot for backup.

	my $vm_view = shift;

	my $snap_moref;

	eval {
	$snap_moref = $vm_view->CreateSnapshot(
			name => "Unitrends_backup",
			description => "Created for backup to Unitrends DPU. (temporary snapshot)",
			memory => 0,
			quiesce => 0 );
	};
	if ($@) {
		if (ref($@) eq 'SoapFault') {
			dbg_log("Error creating temporary snapshot: " . $@);
		}

		return;
	}

	eval {
		my $snap_view = Vim::get_view(mo_ref => $snap_moref);
		$snap_view->RemoveSnapshot(removeChildren => 0);
	};
	if ($@) {
		if (ref($@) eq 'SoapFault') {
			dbg_log("Error removing temporary snapshot: " . $@);
			# TODO: should parse faults like other RemoveSnapshot calls.
		}
	}
}

sub snapshots_exist ($) {
	# Return 0 if no snapshots exist, 1 if some do.

	my $vm_view = shift;

	if (defined $vm_view->snapshot) {
		dbg_log("Snapshots do exist on this VM.");
		return 1;
	} else {
		dbg_log("Snapshots do not currently exist on this VM.");
		return 0;
	}
}

sub independent_disks_present ($) {

	my $vm_view = shift;

	my $devices = $vm_view->config->hardware->device;

	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			if( $_->{'backing'}->diskMode ne 'persistent') {
				# Independent Disks have the value 'independent_persistent'
				# or 'independent_nonpersistent'
				# NOTE: Physical RDMs are independent disks too
				dbg_log("Located an Independent Disk.");
				return 1;
			}
		}
	}

	return 0;

}

sub physical_rdm_disks_present ($) {
	# Only use this call during backup, it's somewhat expensive.
	my $vm_view = shift;

	my $devices = $vm_view->config->hardware->device;

	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			if ($_->{'backing'}->isa('VirtualDiskRawDiskMappingVer1BackingInfo')) {
				if ( ($_->backing->{'compatibilityMode'}) eq "physicalMode") {
					dbg_log("Located a physical RDM disk.");
					return 1;
				}
			}
		}
	}

	return 0;
}

sub device_is_on_nfs ($) {
	my $device = shift;

	my $datastore = $device->backing->datastore;
	my $ds_view = Vim::get_view(mo_ref => $datastore);
	my $ds_info = $ds_view->{'info'};
	if (defined $ds_info) {
		if ($ds_info->can('nas')) {
			if ($ds_info->nas->type eq 'NFS') {
				dbg_log("Device " . $device->key . " is on NFS.");
				return 1;
			}
		}
	}

	if (defined ($device->backing->parent)) {
		my $pds_view = Vim::get_view(mo_ref => $device->backing->parent->datastore);
		my $pds_info = $pds_view->{'info'};
		if (defined $pds_info) {
			if ($pds_info->can('nas')) {
				if ($pds_info->nas->type eq 'NFS') {
					dbg_log("Device " . $device->key . " is on NFS.");
					return 1;
				}
			}
		}
	}

	return 0;
}

sub nfs_disks_present ($) {
	my $vm_view = shift;

	my $devices = $vm_view->config->hardware->device;

	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			if (device_is_on_nfs($_)) {
				dbg_log("Located an NFS disk.");
				return 1;
			}
		}
	}

	return 0;
}

sub configure_change_tracking ($$$$$$) {
	# enable change tracking if it is disabled

	my $vm_moref = shift;
	my $is_full = shift;
	my $xmlroot = shift;
	my $warning_ref = shift;
	my $retry = shift;
	my $promote_next_backup_because_cbt_off_ref = shift;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $has_physical_rdm_disks = physical_rdm_disks_present($vm_view);
	my $has_nfs_disks = nfs_disks_present($vm_view);

	if (not $vm_view->capability->changeTrackingSupported) {
		# Bug 18622 make sure changeTrackingEnabled is
		# off when changeTrackingSupported is false
		disable_change_tracking($vm_view);
		dbg_log("This VM does not support change tracking.");
		return 0;
	}

	if ($vm_view->capability->changeTrackingSupported) {
		if ($vm_view->config->changeTrackingEnabled) {
			dbg_log("Change Block Tracking (CBT) is already ENABLED on this VM.");
			if (snapshots_exist($vm_view) == 0) {
				# No snapshots currently exist on this VM

				if ($is_full) {

					if ($retry) {
						dbg_log("(RETRYING) Performing a FULL Backup of a VM with NO Snapshots.\n" .
							"Was not able to successfully get CBT information.\n" .
							"Now, performing stun-unstun cycle.");

						# On a "RETRY", disable and re-enable if we're doing a full backup
						# and there are no snapshots.
						# This is meant to clean up the issue where change tracking was
						# enabled while snapshots already existed.

						# Do not touch the change tracking settings if RDM disks are present.

						dbg_log("Temporarily disabling change tracking.");
						disable_change_tracking($vm_view);

						dbg_log("Using snapshot create/delete cycle to stun VM.");
						snapshot_stun_cycle($vm_view);

						dbg_log("Re-enabling change tracking.");
						enable_change_tracking($vm_view, $xmlroot);

						#dbg_log("Using snapshot create/delete cycle to stun VM.");
						#snapshot_stun_cycle($vm_view);	# This should not be required

					} else {
						dbg_log("Performing a FULL Backup of a VM with NO Snapshots. " .
							"Not performing the stun-unstun cycle.");
					}

				} else {
					dbg_log("Performing a Full Backup? ===> $is_full");
					dbg_log("VM has one or more pRDMS? ===> $has_physical_rdm_disks");
				}

			} else {
				dbg_log("Not reconfiguring Change Block Tracking (CBT) because snapshots exist.");
			}

			return 1;
		} else {
			dbg_log("Change Block Tracking (CBT) is currently DISABLED on this VM.");

			if (!$is_full) {
				# CBT is currently disabled on the VM and the request is to perform
				# an Incremental/Differential Backup. Can't do it.
				# FAIL current Backup and promote the next one to a Full.

				$$promote_next_backup_because_cbt_off_ref = 1;
				return;
			}

			# Performing a Full Backup
			if (snapshots_exist($vm_view) == 1) {
				# Snapshot/s exist/s currently on this VM
				# Enabling Change Block Tracking (CBT) while snapshots exist can cause problems.

				# perform a non-sparse full backup
				$$warning_ref = $$warning_ref .
					"Warning: Not enabling change tracking because snapshots exist.\n" .
					"Backups will contain all blocks on the VM's disk(s).\n"           . 
					"To enable optimized (sparse) backups, remove all snapshots\n"     .
					"from the VM and then perform a Full backup.\n"                    .
					"Snapshots may be created after this has been done.\n";

				dbg_log("Not enabling change tracking because snapshots exist.\n" .
					"Backup will contain all the blocks and will be the size of the disk/s.");

			} else {
				# No snapshots exist, enable change tracking.
				enable_change_tracking($vm_view, $xmlroot);
			}

			return 0;
		}
	} else {
		dbg_log("Change Block Tracking (CBT) is not supported on this ESX server.");
	}
}

sub handle_snapshot_error {
	# Die with message for errors that can occur with CreateSnapshot()
	# If $may_retry is 1, return 1 for errors that indicate that
	# a second attempt with memory snapshotting enabled, or quiescing
	# disabled, may succeed.

	my $may_retry = shift;  # 0 = always die, 1 = return 1 if retry is possible
	my (@error) = shift;    # Pass $@ by ref.

	my $err_hdr = "Error creating snapshot of VM";

	if (ref($@) eq 'SoapFault') {
		if (ref($@->detail) eq 'RestrictedVersion') {
			die "$err_hdr: This host is not licensed to perform this operation.\n";
		}
		if (ref($@->detail) eq 'ApplicationQuiesceFault') {
			if ($may_retry) {
				# Indicate caller should retry w/o quiescing
				return 1;
			} else {
				die "$err_hdr: Application quiescing failed.\n";
			}
		}
		if (ref($@->detail) eq 'FilesystemQuiesceFault') {
			if ($may_retry) {
				# Indicate caller should retry w/o quiescing
				return 1;
			} else {
				die "$err_hdr: Filesystem quiescing failed.\n";
			}
		}
		# PASSTHROUGH is added to print the '=' correctly in the KB link
		if (ref($@->detail) eq 'FileTooLarge') {
			die "$err_hdr:\n" .
				"File is larger than the maximum size supported by datastore.\n" .
				"PASSTHROUGHPlease see http://support.unitrends.com/ikm/questions.php?questionid=22 for possible steps to resolve this problem.\n";
		}
		if (ref($@->detail) eq 'GenericVmConfigFault') {
			if ($may_retry) {
				# Indicate caller should retry w/o quiescing
				return 1;
			} else {
				die "$err_hdr: GenericVmConfigFault: " . $@->detail->reason . "\n" .
				"PASSTHROUGHPlease see http://support.unitrends.com/ikm/questions.php?questionid=1449 for possible steps to resolve this problem.\n";
			}
		}
		if (ref($@->detail) eq 'HostCommunication') {
			die "$err_hdr: An error occurred while communicating with the remote host.\n";
		}
		if (ref($@->detail) eq 'HostNotConnected') {
			die "$err_hdr: Unable to communicate with remote host.\n";
		}
		if (ref($@->detail) eq 'InvalidName') {
			die "$err_hdr: The snapshot name is invalid.\n";
		}
		if (ref($@->detail) eq 'InvalidDiskFormat') {
			die "$err_hdr: The disk format is either corrupted or not supported.\n";
		}
		if (ref($@->detail) eq 'InvalidState') {
			die "$err_hdr: The operation is not allowed in the current state.\n";
		}
		if (ref($@->detail) eq 'MemorySnapshotOnIndependentDisk') {
			die "$err_hdr: Memory snapshot requested for VM with an independent disk.\n";
		}
		if (ref($@->detail) eq 'MultipleSnapshotsNotSupported') {
			die "$err_hdr: This VM only supports a single snapshot at a time.\n";
		}
		if (ref($@->detail) eq 'NoPermission') {
			die "$err_hdr: Permission to perform this operation was denied: " . $@->detail->privilegeId . " permission required.\n";
		}
		if (ref($@->detail) eq 'NotSupported') {
			die "$err_hdr: Host does not support snapshots.\n";
		}
		if (ref($@->detail) eq 'SnapshotDisabled') {
			die "$err_hdr: Snapshots are disabled on the virtual machine.\n";
		}
		if (ref($@->detail) eq 'SnapshotIncompatibleDeviceInVm') {
			die "$err_hdr: An incompatible device exists on the virtual machine:\n" . $@->fault_string . "\n";
		}
		if (ref($@->detail) eq 'SnapshotLocked') {
			die "$err_hdr: The virtual machine's snapshot is locked.\n";
		}
		if (ref($@->detail) eq 'SnapshotNoChange') {
			die "$err_hdr: The virtual machine's state has not changed since the previous snapshot was created.\n";
		}
		if (ref($@->detail) eq 'SystemError') {
			# Often "Protocol Error from VMX"

			if ($may_retry) {
				return 1;
			} else {
				die "$err_hdr: System Error.\n" . $@;
			}
		}
		if (ref($@->detail) eq 'TaskInProgress') {
			if ($may_retry) {
				# This fault is sometimes returned when VSS has problems -- should retry.

				# Sleep for a short period of time in case this was a legitimate fault
				# caused by another task running.
				sleep 30;

				return 1;
			} else {
				die "$err_hdr: A task is already in progress.\n";
			}
		}
		if (ref($@->detail) eq 'Timedout') {
			if ($may_retry) {
				# Indicate to caller that operation should be retried.
				return 1;
			} else {
				die "$err_hdr: Operation timed out.\n";
			}
		}
		if (ref($@->detail) eq 'TooManySnapshotLevels') {
			die "$err_hdr: Too many snapshots exist.\n";
		}

		# Generic fault containers below, most faults should be caught by something above.

		if (ref($@->detail) eq 'SnapshotFault') {
			# Should be extended by one of the above faults
			die "$err_hdr: A snapshot operation cannot be performed. (SnapshotFault)\n";
		}

		if (ref($@->detail) eq 'VmConfigFault') {
			die "$err_hdr: VmConfigFault\n" . $@ . "\n";
		}

		if (ref($@->detail) eq 'FileAlreadyExists') {
			die "$err_hdr: FileAlreadyExists Fault.\n" .
			"PASSTHROUGHPlease see http://support.unitrends.com/ikm/questions.php?questionid=1439 " .
			"for possible steps to resolve this problem.\n" . $@ . "\n";
		}
	}

	# Failure was not caught above.
	die "$err_hdr: " . $@ .  "\n";
}

# This will create the snapshot using the VMware API as per the parameters passed in.
# If memory and quiesce both are passed in as zero/false, and if the snapshot operation
# fails, then, it will die

sub create_snapshot_api {

	my $vm_view = shift;
	my $snap_name = shift;
	my $description = shift;
	my $memory_api_option = shift;
	my $quiesce_api_option = shift;

	my $snap_moref;
	my $snap_return_value = 1;

	eval {
		dbg_log("Calling CreateSnapshot with memory = $memory_api_option," .
				" quiesce = $quiesce_api_option");
		$snap_moref = $vm_view->CreateSnapshot(
			name => $snap_name,
			description => $description,
			memory => $memory_api_option,
			quiesce => $quiesce_api_option );
	};
	if ($@) {
		dbg_log("Failed creating snapshot with memory = $memory_api_option," .
				" quiesce = $quiesce_api_option");
		dbg_log("Error from VMware API: $@");

		if ( ($memory_api_option eq 0) and ($quiesce_api_option eq 0) ) {
			# Will die
			handle_snapshot_error(0, \$@);
		}

		$snap_return_value = 0;
	}

	return ($snap_moref, $snap_return_value);

}

# Will always try to first snapshot the VM with memory = 1 and quiesce = 1
# If that fails, snapshot the VM with memory = 0 and quiesce = 0
sub helper_snap_operations {

	my $vm_view = shift;
	my $snap_name = shift;
	my $description = shift;
	my $has_independent_disks = shift;
	my $warning_ref = shift;
	my $info_ref = shift;

	my $quiesce_api_option = 1;
	my $memory_api_option = 1;
	my $snap_retval = 1;
	my $snap_moref;

	if ($has_independent_disks) {
		dbg_log("VM has one or more independent disks. Will directly" .
			" create a snapshot with memory = 0 and quiescing = 0.");
		goto NO_QUIESCE;
	}

	($snap_moref, $snap_retval) = create_snapshot_api($vm_view, $snap_name,
						$description, $memory_api_option, $quiesce_api_option);

	if (!$snap_retval) {
		# Snapshot with quiescing as well as memory failed
		# So, try snapping the VM with both as false and put
		# a warning in the Backup if snapshot is now successful

		NO_QUIESCE:
		$snap_retval = 1;
		$quiesce_api_option = 0;
		$memory_api_option = 0;

		($snap_moref, $snap_retval) = create_snapshot_api($vm_view, $snap_name,
						$description, $memory_api_option, $quiesce_api_option);

		$$warning_ref = $$warning_ref . "\n" .
				"This backup encountered a snapshot timeout and was " .
				"performed with a snapshot without I/O quiescing i.e., " .
				"the snapshot was performed with crash-consistent setting.\n";

	} else {
		dbg_log("Successfully created snapshot with memory = $memory_api_option " .
			"and quiesce = $quiesce_api_option");

		$$info_ref = $$info_ref . "\n" .
				"Snapshot for this Backup was created with " .
				"application-consistent setting.\n";
	}

	return $snap_moref;
}

sub create_snapshot ($$$$$$$$$) {

	# returns: moref of snapshot or dies with a failure message

	my $vm_moref = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $warning_ref = shift;
	my $err = shift;        # Error returned from do_guest_vss_snapshot
	my $has_independent_disks = shift;
	my $quiesce = shift;
	my $appaware = shift;
	my $info_ref = shift;

	my $memory_api_option;
	my $quiesce_api_option;
	my $snap_retval = 1;

	dbg_log("Value of err is $err.");
	dbg_log("Independent Disk/s present? => $has_independent_disks.");

	# Need to call CreateSnapshot on a view, not a moref.
	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snap_moref;  # New moref of created snapshot.

	my $snap_name = "Unitrends_backup";
	my $timestamp = strftime("%H:%M %Z %a %b %e %Y", localtime);
	my $description = "Created for backup to Unitrends DPU: " . $timestamp;

	# From release 9.1.0, we are supporting an additional quiesce option per VM
	# The VM being Backed up will be snapshotted according to that option

	# Crash-consistent setting for the snapshot
	if ( ($appaware eq "NO") and ($quiesce eq "NO") ) {
		$memory_api_option = 0;
		$quiesce_api_option = 0;
		($snap_moref, $snap_retval) = create_snapshot_api($vm_view, $snap_name,
							$description, $memory_api_option, $quiesce_api_option);

		if ($snap_retval) {
			$$info_ref = $$info_ref . "\n" .
					"Snapshot for this Backup was created with" .
					" crash-consistent setting.\n";
		}
	}

	# App-aware setting for the snapshot
	# If app-aware was successful, then, create a crash-consistent snapshot
	# If app-aware failed, then, attempt to create a application-consistent snapshot
	if ( ($appaware eq "YES") and ($quiesce eq "NO") ) {
		if ($err == 0) {
			# $err will be set to zero when appaware processing was successful

			$memory_api_option = 0;
			$quiesce_api_option = 0;
			dbg_log("Guest snapshot operation was successful. Will call " .
				"CreateSnapshot with memory = $memory_api_option and " .
				"quiesce = $quiesce_api_option.");

			($snap_moref, $snap_retval) = create_snapshot_api($vm_view, $snap_name,
							$description, $memory_api_option, $quiesce_api_option);

			if ($snap_retval) {
				$$info_ref = $$info_ref . "\n" .
						"Snapshot for App-aware processing was successful.\n"
			}

		} else {
			# Appaware processing failed. Try snapshotting with quiesce as true.
			# However, complete the Backup with a warning, indicating that appaware
			# processing failed

			dbg_log("App-aware processing failed. Will first try to snapshot " .
					" the VM with quiesce = 1");
			$$warning_ref = $$warning_ref . "\n" .
					"App-aware snapshots enabled, but failed to create " .
					"guest snapshot.\n";
			goto QUIESCE_ONLY;
		}
	}

	# Application-consistent setting for the snapshot
	if ( ($appaware eq "NO") and ($quiesce eq "YES") ) {

		QUIESCE_ONLY:
		$memory_api_option = 0;
		$quiesce_api_option = 1;
		($snap_moref, $snap_retval) = create_snapshot_api($vm_view, $snap_name,
							$description, $memory_api_option, $quiesce_api_option);

		if (!$snap_retval) {
			# Anytime a snapshot with quiescing fails, attempt to
			# snapshot the VM first with memory = 1 AND quiesce = 1
			# 
			# If that also fails, then, snapshot the VM with
			# memory = 0 AND quiesce = 0
			#
			# If that too fails, then fail the Backup

			$memory_api_option = 1;
			$quiesce_api_option = 1;
			$snap_moref = helper_snap_operations($vm_view, $snap_name, $description,
						$has_independent_disks, \$warning_ref, \$info_ref);
		
		} else {
			dbg_log("Successfully created snapshot with memory = $memory_api_option " .
				"and quiesce = $quiesce_api_option");
			$$info_ref = $$info_ref . "\n" .
					"Snapshot for this Backup was created with" .
					" application-consistent setting.\n";
		}
	}

	if ( ($appaware eq "YES") and ($quiesce eq "YES") ) {
		die "Can't perform a Backup with both - application aware processing" .
			" as well as quiescing.\n";
	}

	return $snap_moref;

}

sub transport_mode {

	my $xmlroot = shift;

	my @contents;

	if (open(my $CONFIG_FILE, '<', '/usr/bp/bin/vprotect/mode.conf')) {

		while(<$CONFIG_FILE>) {
			chomp;
			@contents = split('=', $_);
			if (defined $contents[1]) {
				if ($contents[1] eq "true") {
					dbg_log("Using best Transport Mode for Backup");
					$xmlroot->appendTextChild('bestTransportMode', 'true');
				} elsif ($contents[1] eq "false") {
					dbg_log("Using NBD Transport Mode for Backup");
					$xmlroot->appendTextChild('bestTransportMode', 'false');
				} else {
					dbg_log("Unable to determine Transport Mode for Backup");
					dbg_log("Contents:" . Dumper($contents[1]));
					die "Unable to determine transport mode for Backup from configuration file\n";
				}
			}
		}

		close($CONFIG_FILE)
			or warn "Unable to close configuration file: $!";

	} else {
		# Unable to open the mode.conf file for some reason
		# Log the message, use the default/best transport mode, but don't fail the Backups
		dbg_log("Unable to open the mode.conf file - $!");
		dbg_log("Using the default/best transport mode for Backup.");
		$xmlroot->appendTextChild('bestTransportMode', 'true');
	}

}

##
## Kas 9/22/2012 : Bug 14809
## This method will search for the Unitrends_backup
## snapshot in the childSnapshotList moref.
##
sub find_unitrends_snap ($); 
sub find_unitrends_snap ($) {
	my $snap_tree = shift;
	my $snapRet;
	
	dbg_log("In find_unitrends_snap: " . ref($snap_tree));
	if ($snap_tree->childSnapshotList) {
		dbg_log("snap_tree defined");
		foreach (@{$snap_tree->childSnapshotList}) {
			dbg_log("snap_tree entry");
			$snapRet = find_unitrends_snap($_);
		}
	}

	if ($snap_tree->name eq "Unitrends_backup") {
		dbg_log("Found unitrends_snap");
		$snapRet = Vim::get_view(mo_ref => $snap_tree->snapshot);
	} else {
		dbg_log("Found leaf, not unitrends snap.");
	}

	return $snapRet;
}

sub get_disk_info ($$$$$$$$$$) {
	# params: moref for VM, xml doc, xml root node
	my $vm_moref = shift;
	my $snap_moref = shift;
	my $change_id_hash_ref = shift;
	my $keys_array = shift;
	my $is_template = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $is_full = shift;
	my $promote_next_ref = shift;
	my $snapshots_exist = shift;

	dbg_log("Entering get_disk_info");

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my @disks;
	my $snap;
	my $devices;
	my $disks_backup_size = 0;

	my $reset_cbt;

	my @keys_array_ref = @$keys_array;
	
	dbg_log("keys_array is @keys_array_ref");

	my $disks_xmlelem = $xmldoc->createElement('disks');

	if ($is_template eq "NO") {
		##
		## Kas 9/22/2012 : Bug 14809
		## Typically the snapshot->currentSnapshot identifies the
		## newest snapshot.  If that entry is empty, then look for
		## the Unitreds_backup snapshot in the snapshot list.
		## currentSnapshot has been seen empty on ESXi 5.1.
		##
		my $snap_moref_t = $vm_view->snapshot->currentSnapshot;
		if (!defined $snap_moref_t) {
			dbg_log("get_disk_info: No currentSnapShot reference. Will search for UT snapshot.");
			if (@{$vm_view->snapshot->rootSnapshotList}) {
				foreach (@{$vm_view->snapshot->rootSnapshotList}) {
					$snap = find_unitrends_snap( $_ );
					if (defined($snap)) {
						last;
					}
				}
			}
		} else {
			dbg_log("get_disk_info: Using currentSnapshot reference.");
			$snap = Vim::get_view(mo_ref => $snap_moref_t);
		}
		$devices = $snap->config->hardware->device;
	} else {
		$devices = $vm_view->config->hardware->device;
	}

	my $flag = 0;
	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			$flag = 0;
			my $current_disk_backup_size = 0;
			# Bug 16541 - Exclude a pRDM, when 'empty' exclude argument is received to vprotect
			my $backing_comp_mode = $_->backing->{'compatibilityMode'};
			if (defined $backing_comp_mode) {
				if ($backing_comp_mode eq 'physicalMode') {
					foreach my $iterator (@keys_array_ref) {
						if ($iterator eq 'empty') {
							dbg_log("Exclude argument is - $iterator");
							dbg_log("A VM with Physical RDM found and no arguments were passed to exclude. " .
								"Skipping this disk.");
							$flag = 1;
						}
					}
				}
			}

			if ($flag == 1) {
				next;
			}

			my $disk = {};
			my $disk_xmlelem = $xmldoc->createElement('disk');

			$disk->{'key'} = $_->key;

			dbg_log("Dealing with disk with key " . $_->{'key'});

			$disk_xmlelem->appendTextChild('key', $_->{'key'});
			$disk_xmlelem->appendTextChild('capacityInKB', $_->{'capacityInKB'});
			$disk_xmlelem->appendTextChild('fileName', $_->backing->{'fileName'});
			$disk_xmlelem->appendTextChild('label', $_->{'deviceInfo'}{'label'});

			# Excluding Disks. Won't get the changed block areas for these disks
			foreach my $iterator (@keys_array_ref) {
				if (($_->{'key'} eq $iterator) || ($_->{'backing'}->diskMode ne 'persistent')) {
					if ($iterator == 'empty') {
						dbg_log("Skipping an independent disk by default.");
					} else {
						dbg_log("Skipping disk with key $iterator.");
					}
					dbg_log("Skipping disk with key $iterator");
					$flag = 1;
					last;
				}
			}
			
			if ($flag == 1)	{
				next;
			}

			###
			# Get the changed disk areas	
			###
			eval {
				($reset_cbt, $current_disk_backup_size) = get_changed_blocks_for_disk($vm_moref, $snap_moref, $_->{'key'}, $change_id_hash_ref->{$_->{'key'}}, $is_template, $xmldoc, $disk_xmlelem, $is_full, $promote_next_ref, $_->{'capacityInKB'}, $snapshots_exist);

				if ($reset_cbt == 33) {
					dbg_log("\tResetting CBT since got an error related to CBT.");
					$xmlroot->removeChild($disks_xmlelem);
				}

			};
			if ($@) {
				die "Failed to get changed blocks for disk. Remove any snapshots that may exist for the guest: " . $@ . "\n";
			}

			if ($reset_cbt == 33) {
				dbg_log("\tWill perform a retry to get CBT information.");
				last;	# Exit 'foreach (@$devices)', since have to retrieve CBT
					# information _again_ for ALL disks
			} else {
				dbg_log("Successfully retrieved CBT information.");
				push @disks, $disk;
				$disks_xmlelem->addChild($disk_xmlelem);
				$disks_backup_size = $disks_backup_size + $current_disk_backup_size;
			}
		}
	}

	if ($reset_cbt == 33) {
		dbg_log("\tResetting CBT since got an error related to CBT.");
		return $reset_cbt;
	} else {
		my $backup_size_in_mb = $disks_backup_size/1024/1024;
		dbg_log("Current Backup size in MBs is - $backup_size_in_mb");
		$xmlroot->appendTextChild('backupSizeMb', $backup_size_in_mb);

		$xmlroot->addChild($disks_xmlelem);
		dbg_log("Returning all disks");
		return @disks;
	}

}

sub process_changed_disk_areas {
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $xmlelem = shift;
	my $key = shift;
	my $result = shift;

	my $size;
	my %result_hash = %$result;

	if (defined($result_hash{'changedArea'})) {
		# changed blocks were found
		my $num_areas = $result_hash{'changedArea'};  # arrayref
		my @deref_areas = @$num_areas;

		my $area_count = scalar @deref_areas;
		$xmlelem->appendTextChild('num_areas', $area_count);

		foreach (@deref_areas) {
			my $xmlblockelem = $xmldoc->createElement('changed_disk_area');
			$xmlblockelem->appendTextChild('length', $_->{'length'});
			$xmlblockelem->appendTextChild('start', $_->{'start'});
			$xmlelem->addChild($xmlblockelem);
		}
		
		$size = $result_hash{'length'};
	} else {
		# the whole disk is empty (it has never been used)
		$size = 0;
	}

	return $size;

}

sub get_change_id_from_snap ($$) {
	my $snap_moref = shift;
	my $device_key = shift;

	dbg_log("Entering get_change_id_from_snap...");
	my $snap_view = Vim::get_view(mo_ref => $snap_moref);

	my @devices = $snap_view->config->hardware->device;
	my $i = 0;
	while ($i < scalar $devices[0]) {
		# TODO: examine this further
		if ($devices[0][$i]->key != $device_key) {
			$i = $i + 1;
			next;
		}
		return $devices[0][$i]->backing->changeId;
	}
}


sub get_changed_blocks_for_disk ($$$$$$$$$$$) {
	my ($vm_moref, $snap_moref, $devicekey, $changeid, $is_template, $xmldoc, $xmlroot, $is_full, $promote_next_ref, $capacityInKB, $snapshots_exist) = @_;

	dbg_log("Entering get_changed_blocks_for_disk.");

	my $capacityInBytes = $capacityInKB * 1024;

	dbg_log("Disk capacityInKB is $capacityInKB");
	dbg_log("Disk capacityInBytes is $capacityInBytes");
	dbg_log("Snapshots Exist? ===> $snapshots_exist.");

	my $disk_backup_size = 0;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	# Call QCDA only for a VM Backup; not for a Template Backup
	if ($is_template eq "NO") {
		# Check vm_view->capabilities so we know if we can call queryChangedDiskAreas
		if (!$vm_view->capability->changeTrackingSupported) {
			# Server doesn't know about this capability, so it's not supported.
			$xmlroot->appendTextChild('changeTrackingEnabled', '0');
			return (1,$capacityInBytes);
		}
		if (!$vm_view->config->changeTrackingEnabled) {
			# Server has this capability, but the VM doesn't.
			$xmlroot->appendTextChild('changeTrackingEnabled', '0');
			return (1,$capacityInBytes);
		}

		my $nfs_stop = 0;

		# Is deviceKey an RDM disk?
		# If so, skip QCDA, backup whole disk
		my $devices = $vm_view->config->hardware->device;
		my $devCapacity = 0;

		foreach my $dev (@$devices) {
			if ($dev->key == $devicekey) {
				#Bug 15878. Get disk capacity to be used in getting all change-block-areas.
				$devCapacity = ($dev->capacityInKB * 1024);

				if ($is_full and $dev->backing->isa('VirtualDiskRawDiskMappingVer1BackingInfo')) {
					dbg_log("Found RDM disk $devicekey, skipping QCDA() for this device.");
					$changeid = get_change_id_from_snap($snap_moref, $devicekey);
					dbg_log("Found RDM changeId " . $changeid . " from snapshot.");
					if (defined $changeid) {
						$xmlroot->appendTextChild('changeId', $changeid);
					}
					$xmlroot->appendTextChild('changeTrackingSkipped', '1');
					$xmlroot->appendTextChild('changeTrackingEnabled', '1');   # Skip QCDA() later.
					return (1,$devCapacity);
				}

				# Is this device on an NFS datastore?
				if (device_is_on_nfs($dev) and $is_full) {
					# Don't call QueryChangedDiskAreas if the disk is on NFS and we are doing a full.
					# Only supported for diff/incr backups.

					dbg_log("Skipping QueryChangedDiskAreas() since device " . $dev->key . " is on NFS, and this is a full backup.");
					# Report change tracking as enabled, but don't use it now
					# This is so the changeId gets recorded.
					$xmlroot->appendTextChild('changeTrackingEnabled', '1');

					# Record that we aren't using change tracking for this particular backup.
					$xmlroot->appendTextChild('changeTrackingSkipped', '1');

					# Will backup the complete disk
					$disk_backup_size = $devCapacity;

					$nfs_stop = 1;
				} else {
					$xmlroot->appendTextChild('changeTrackingSkipped', '0');
				}

				last;  # Only one device will match anyway, stop looking
			}
		}

		my $change_id = get_change_id_from_snap($snap_moref, $devicekey);
		if (defined $change_id) {
			$xmlroot->appendTextChild('changeId', $change_id);
		}

		if ($nfs_stop) {
			# Don't actually use QueryChangedDiskAreas
			return (1,$capacityInBytes);
		}

		if (!defined $changeid and $is_full) {
			$changeid = "*";
		}

		if (!defined $changeid and !$is_full) {
			# This used to set changeid = "*", now we want to fail the backup, since
			# this only happens if a disk is queried which we do not have a change id for.
			$$promote_next_ref = 1;
			die "Disk $devicekey was added to this VM -- a full backup is required.\n";  # TODO message
		}

		dbg_log("Calling QueryChangedDiskAreas with snapshot: " . $snap_moref->value .
				", deviceKey: " . $devicekey . ", changeId: " . $changeid);

		my $startOffset = 0;
		my $areaSize = 0;
		my $processChanges = 1;

		eval {
			my @result;
			my $xmlElem;
			#Bug 16009. Only add changed_disk_areas element if there changes to backup
			#for the current disk. 'addElem' determines if the element is to be added.
			my $addElem = 0;

			#Bug 15878. New loop.
			#Call QueryChangedDiskAreas until we span the entire disk capacity.
			#May require more than 1 call to gather all change-block-areas.
			while  ($processChanges == 1) {
				@result = $vm_view->QueryChangedDiskAreas(
					snapshot => $snap_moref,
					deviceKey => $devicekey,
					startOffset => $startOffset,
					changeId => $changeid);

				if ($startOffset == 0) {
					$xmlElem = $xmldoc->createElement('changed_disk_areas');
				}

				$areaSize = process_changed_disk_areas($xmldoc, $xmlroot, $xmlElem, $devicekey, @result);
				dbg_log("QueryChangedDiskAreas returns start " . $startOffset . " : Size " . $areaSize);

				if ($areaSize > 0) {
					$startOffset += $areaSize;
					$addElem = 1;
				} else {
					$processChanges = 0;
				}

				if ($startOffset >= $devCapacity) {
					$processChanges = 0;
				}	
			}

			if ($addElem == 1) {
				use bigint;

                                my @starts = ();
                                my @lengths = ();

				foreach my $elem ($xmlElem->findnodes('changed_disk_area')) {
                                    dbg_log("Original start: " . $elem->findvalue('start') . "\tlength: " . $elem->findvalue('length'));
                                    my $start_value = $elem->findvalue('start');
                                    my $lower = ($start_value / 1048576)*1048576;
					#We need to find the actual position, and figure out the next boundary
					#on that position, then calculate the new length, we just can't
					#arbitrarily extend it out to the next one meg boundary.
					#find the length in the physical file.
                                    my $length_value = $elem->findvalue('length') + $start_value;
                                    my $upper = $length_value;
					#Now calculate to the next 1 meg boundary for the physical
					#locate in the file.
                                    if (($length_value % 1048576) gt 0) {
                                        $upper = (1048576-($length_value % 1048576))+$length_value;
                                    }
					#now calculate the actual true length we need.
				    $upper = $upper - $lower;
                                    dbg_log("Modified start: " . $lower . "\tlength: " . $upper);
				
				    if ($lower < $capacityInBytes) {
                                    	push(@starts, $lower);
					if (($lower + $upper) > $capacityInBytes) {
						# This means that we are trying to access something out of bounds
						$upper = $capacityInBytes - $lower;
						dbg_log("Modified start: " . $lower . "\tlength: " . $upper . " (last)");
					}
                                    	push(@lengths, $upper);
				    }
                                }

                                dbg_log("Cnt0 = " . scalar(@starts));
                                
                                for (my $i=0; $i < scalar(@starts); $i++) {
                                    dbg_log("Step1 start: " . $starts[$i] . "\tlength: " . $lengths[$i]);
                                }

                                
                                my $prevStart = 0;
                                my $prevLength = 0;
                                my $currStart = 0;
                                my $currLength = 0;
                                
                                my @newStarts = ();
                                my @newLengths = ();

                                for (my $i=0; $i < scalar(@starts); $i++) {
                                    if ($i==0) {
                                        $prevStart = $starts[$i];
                                        $prevLength = $lengths[$i];
                                        next;
                                    }
                                    $currStart = $starts[$i];
                                    $currLength = $lengths[$i];
                                    if ($prevStart != $currStart) {
                                        push(@newStarts, $prevStart);
                                        push(@newLengths, $prevLength);
                                        $prevStart = $currStart;
                                        $prevLength = $currLength;
                                    } else {
                                        if ($prevLength < $currLength) {
                                            $prevLength = $currLength;
                                        }
                                    }
                                }
                                push(@newStarts, $prevStart);  # To take care of the last one
                                push(@newLengths, $prevLength);
                                
                                dbg_log("Cnt1 = " . scalar(@newStarts));
                                for (my $i=0; $i < scalar(@newStarts); $i++) {
                                    dbg_log("Step2 start: " . $newStarts[$i] . "\tlength: " . $newLengths[$i]);
                                }


                                my $revised_xmlElem = $xmldoc->createElement('changed_disk_areas');
                                $revised_xmlElem->appendTextChild('num_areas', scalar(@newStarts));

                                for (my $i=0; $i < scalar(@newStarts); $i++) {
                                    my $xmlblockelem = $xmldoc->createElement('changed_disk_area');
                                    $xmlblockelem->appendTextChild('start', $newStarts[$i]);
                                    $xmlblockelem->appendTextChild('length',$newLengths[$i]);
                                    $revised_xmlElem->addChild($xmlblockelem);
					$disk_backup_size = $disk_backup_size + $newLengths[$i];
                                }
                                $xmlroot->addChild($revised_xmlElem);
                            }

			$xmlroot->appendTextChild('changeTrackingEnabled', '1');
                    };
		if ($@) {
			# QueryChangedDiskAreas (QCDA) returned a failure

			if (ref ($@) eq 'SoapFault') {
				my $params = "snapshot: " . Dumper($snap_moref) .
					" deviceKey: " . $devicekey .
					" startOffset: 0 " .
					" changeId : " . $changeid;

				dbg_log("Fault type: " . ref ($@));
				dbg_log("SoapFault type: " . ref($@->detail));

				if (ref($@->detail) eq 'FileFault') {
					dbg_log("FileFault:\n" . Dumper($@));
					dbg_log("Params:\n" . $params);

					if ( ($changeid eq "*") and (!$snapshots_exist) ) {
						#die "Error getting changed disk areas, FileFault.\n\n" .
						#$@ . "\n" . $params . "\n";

						# Error getting CBT info on a FULL Backup of a VM with no snapshots
						# Reset the CBT flag on this VM on the fly and perform
						# the Backup using the new CBT information

						# Return a special error code
						dbg_log("\t\tReturning SPECIAL error code.");
						return (33, 0); # TODO Change this if required

					} else {
						$$promote_next_ref = 1;

						die "Error getting changed disk areas: FileFault.\n" . 
						"This issue may be resolved by removing all snapshots and " .
						"performing a Full backup of this VM.\n\n" .
						"PASSTHROUGHPlease see https://unitrends-support.force.com/UnitrendsBackup/s/article/000003170 " .
						"for possible steps to resolve this problem.\n\n" .
						$@ . "\n" . $params . "\n";
					}

				} elsif (ref($@->detail) eq 'NoPermission') {
					dbg_log("NoPermissionFault:\n" . Dumper($@));
					dbg_log("Params: \n" . $params);
					die "Error getting changed disk areas: NoPermissionFault.\n" .
						"This issue may be resolved by adding role permissions for " .
						$@->detail->privilegeId . " in your vCenter server.\n\n";
				}
			}
			die Dumper($@);
		}
	} # End 'if is_template' loop
	else {
		dbg_log("Backup everything for a Template since can't snapshot it.");
		$xmlroot->appendTextChild('changeTrackingEnabled', '0');
		$xmlroot->appendTextChild('changeTrackingSkipped', '1');
	}

	return (1, $disk_backup_size); # Success
}

sub split_change_ids ($) {
	my $change_ids = shift;

	if (!defined $change_ids) {
		return;
	}

	my @fields = split(/:/, $change_ids);

	my %hash;
	my $hashref = \%hash;

	while (scalar @fields > 0) {
		my $a = shift @fields;
		my $b = shift @fields;

		$hash{$a} = $b;
	}

	return $hashref;
}

sub split_keys ($) {
	my $keys = shift;

	my @array1;
	my $arrayref = \@array1;

	if (!defined $keys) {
		return $arrayref;
	}

	my @fields = split(/:/, $keys);

	my $i = 0;
	while (scalar @fields > 0) {
		$array1[$i] = shift @fields;
		$i = $i + 1;
	}

	return $arrayref;
}

sub guest_vss_operation {
	my $command = shift;
	my $vmName = shift;
	my $vmxPath = shift;
	my $server = shift;
	my $hostUser = shift;
	my $hostPasswd = shift;
	my $vmUser = shift;
	my $vmPasswd = shift;
	my $backupStatus = shift;
	my $warning_ref = shift;
	my $hostPath;
	my $hostName;
	my $hostPort;
	my $err;
	my $cmdArgs;
	my $snapstatus;
	my %procinfo;
	#my $toolsConfPath = "ProgramData\\VMware\\VMware tools\\tools.conf";


	($err, $gos_hostHandle) = HostConnect(
				VIX_API_VERSION,
				VIX_SERVICEPROVIDER_VMWARE_VI_SERVER,
				$server,
				$hostPort, 
				$hostUser, 
				$hostPasswd,
				0,
				VIX_INVALID_HANDLE);

	dbg_log("HostConnect: " . $server . " : " . $hostUser . ":  returns " . $err);
	if ($err != 0) {
		my $warnMsg;
		$$warning_ref .= "\n Failed to connect to host for guest vss operations. Incorrect credentials. " . GetErrorText($err);
		return $err;
	}	

	($err, $gos_vmHandle) = VMOpen($gos_hostHandle, $vmxPath);
	dbg_log("VMOpen of " . $vmName . " returns " . $err);
	if ($err != 0) {
		$$warning_ref .= "\n Failed to connect with guest " . $vmName . " for guest vss operations. " . GetErrorText($err); 
		HostDisconnect($gos_hostHandle);
		return $err;
	}

	$err = VMLoginInGuest($gos_vmHandle, $vmUser, $vmPasswd, 0);
	dbg_log("VMLogin returns " . $err);
	if ($err != 0) {
		$$warning_ref .= "\n Failed to login to guest " . $vmName . " for guest vss operations. " . GetErrorText($err);
		HostDisconnect($gos_hostHandle);
		return $err;
	}

	my $winDir;
	my $arch;
	my $exists;

	## Get some important env variables from guest.
	($err, $winDir) = VMReadVariable($gos_vmHandle, VIX_GUEST_ENVIRONMENT_VARIABLE, "WINDIR", 0);
	dbg_log("ReadVarible returns : err " . $err . " : var : " . $winDir);
	if ($winDir eq "") {
		## Super.  Cannot read the %windir% env variable from the guest.
		## To find the Windows directory, scan thru the drive letters until
		## we find it.
		my $vl;
		for $vl ("A".."Z") {
			$winDir = $vl . ":\\Windows";
			($err, $exists) = VMDirectoryExistsInGuest($gos_vmHandle, $winDir);
			if ($exists) {
				dbg_log("WINDIR = " . $winDir);
				last;
			} else {
				$winDir = "";
			}
		}
	}

	## No Windows directory? No problem. Warn the backup an continue without performing the VSS snap.
	if ($winDir eq "") {
		$$warning_ref = $$warning_ref . "\n Failed to find Windows directory in guest. " . $err;
		VMLogoutFromGuest($gos_vmHandle);
		HostDisconnect($gos_hostHandle);
		return 1;
	}

	my $winvolIdx = index($winDir, '\\');
	my $winVol = substr($winDir,0,$winvolIdx+1);

	## Determine the processor architecture so we can run the correct VSS program there.
	($err, $arch) = VMReadVariable($gos_vmHandle, VIX_GUEST_ENVIRONMENT_VARIABLE, "PROCESSOR_ARCHITECTURE", 0);
	dbg_log("ReadVarible returns : err " . $err . " : var : " . $arch);
	if ($arch eq "") {
		my $wowdir;
		$wowdir = $winDir . "\\SysWOW64";
		($err, $exists) = VMDirectoryExistsInGuest($gos_vmHandle, $wowdir);
		if ($exists) {
			dbg_log("Windows\\SysWOW64 EXISTS");
			$arch = "AMD64";
		} else {
			dbg_log("Windows\\SysWOW64 does not EXISTS");
			$arch = "X86";
		}
	}

	## Place the VSS agent in %WinDir%\Windows\Unitrends
	$gos_guestPath = $winDir . "\\Unitrends";
	$gos_guestVssPgmPath = $gos_guestPath . "\\VProtectVss.exe";

	($err, $exists) = VMDirectoryExistsInGuest($gos_vmHandle, $gos_guestPath);
	if (!$exists) {
		## Make the Unitrends directory
		my $mkDirCmd = $winDir . "\\system32\\cmd.exe ";
		$cmdArgs = " /C mkdir " . $gos_guestPath;
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$mkDirCmd,
					$cmdArgs,
					0,
					VIX_INVALID_HANDLE);	
		dbg_log("guest mkdir of : " . $gos_guestPath . " returns " . $err);
		if ($err != 0) {
			$$warning_ref = $$warning_ref . "\n Failed to create temporary Unitrends directory in guest. " . GetErrorText($err);
			VMLogoutFromGuest($gos_vmHandle);
			HostDisconnect($gos_hostHandle);
			return $err;
		}
	}

	## Copy the GetOSver program over and execute to get OS version.
	$gos_osLevel = 6;
	$err = VMCopyFileFromHostToGuest($gos_vmHandle, "/usr/bp/bin/vprotect/GetOsVer.bat", $gos_guestPath . "\\GetOsVer.bat", 0, VIX_INVALID_HANDLE);
	dbg_log("VMCopyFile of : GetOsVer.bat returns " . $err);
	if ($err == 0) {
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
				$winDir . "\\system32\\cmd.exe ",
				"/C " . $gos_guestPath . "\\GetOsVer.bat",
				0,
				VIX_INVALID_HANDLE);	
		$gos_osLevel = $procinfo{'EXIT_CODE'};
		dbg_log("GetOSVersion returns : " . $gos_osLevel);
	}

	if ($command eq "create") {
		## Creating the snapshot

		#if Win 2003 or earlier always run the 32-bit binaries.
		if ($gos_osLevel == 5) {
			$arch = "X86";
		}
		if ($arch eq "AMD64") {
			$hostPath = "/usr/bp/bin/vprotect/VProtectVss64.exe";
		} else {
			$hostPath = "/usr/bp/bin/vprotect/VProtectVss.exe";
		}

		## Copy the VSS program over.
		$err = VMCopyFileFromHostToGuest($gos_vmHandle, $hostPath, $gos_guestVssPgmPath, 0, VIX_INVALID_HANDLE);
		dbg_log("VMCopyFile of : " . $hostPath . " returns " . $err);
		if ($err != 0) {
			$$warning_ref .= "\n Failed to copy vss snapshot server to guest. Cannot continue with snapshot. " . GetErrorText($err);
			VMLogoutFromGuest($gos_vmHandle);
			HostDisconnect($gos_hostHandle);
			return $err;
		}

		## Run the VSS program in server mode first. Leaving it running awaiting commands
		## to create vss snapshots and then later remove them and exit.
		$cmdArgs = " Server  " . $gos_guestPath . "\\VssLogSVR.log";
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$gos_guestVssPgmPath,
					$cmdArgs,
					VIX_RUNPROGRAM_RETURN_IMMEDIATELY,
					VIX_INVALID_HANDLE);	
		dbg_log("RunPgmInGuest " . $gos_guestVssPgmPath . " " . $cmdArgs . " returns " . $err);
	
		if ($err != 0)  {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot server in guest. Cannot create VSS snapshot. " . GetErrorText($err)); 
		} else {
			$cmdArgs = " CreateSnap " . $gos_guestPath . "\\VssLogClnt.log";
			($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
						$gos_guestVssPgmPath,
						$cmdArgs,
						0,
						VIX_INVALID_HANDLE);	
			$snapstatus = $procinfo{'EXIT_CODE'}; 
			if (($err != 0) or ($snapstatus != 0))  {
				## Display Windows hex error code.
				if ($snapstatus == 1) {
					$$warning_ref .= "\n Failed to run vss snapshot create program in guest. Guest VSS snapshots are not supported for guest OS.\n";
				} else {	
					$$warning_ref .= sprintf("\n Failed to run vss snapshot create program in guest. Error 0x%x, VSS program returns error 0x%x\n", $err & 0xFFFFFFFF,  $snapstatus & 0xFFFFFFFF);
				}
			}
		}

		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : error code 0x%x\n",$gos_guestVssPgmPath,$cmdArgs,$err & 0xFFFFFFFF,$snapstatus & 0xFFFFFFFF);

		## Failure to create snapshot is fatal.
		if (($err != 0) or ($snapstatus != 0)) {
			VMLogoutFromGuest($gos_vmHandle);
			HostDisconnect($gos_hostHandle);
			if ($snapstatus != 0) {
				$err = $snapstatus;
			}
			return $err;
		}
	} else {
		## Removing the snapshot
		my %procinfo;
		my $hostLogFile;
		my $guestLogFile;
		$snapstatus = 0;

		if ($backupStatus eq "SUCCESS") {
			$cmdArgs = " ReleaseOk " . $gos_guestPath . "\\VssLogClnt.log";
		} else {
			$cmdArgs = " ReleaseFail " . $gos_guestPath . "\\VssLogClnt.log";
		}
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$gos_guestVssPgmPath,
					$cmdArgs,
					0,
					VIX_INVALID_HANDLE);	
		$snapstatus = $procinfo{'EXIT_CODE'}; 

		if (($err != 0) or ($snapstatus != 0)) {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot release in guest. Error 0x%x : exit code 0x%x\n", $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);
		} 
		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : exit code 0x%x\n", $gos_guestVssPgmPath, $cmdArgs, $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);

		$cmdArgs = " Exit " . $gos_guestPath . "\\VssLogClnt.log";
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$gos_guestVssPgmPath,
					$cmdArgs,
					0,
					VIX_INVALID_HANDLE);	
		dbg_log("RunPgmInGuest (shutdown vss agent) " . $gos_guestVssPgmPath . " " . $cmdArgs . " returns " . $err);
		$snapstatus = $procinfo{'EXIT_CODE'}; 

		if (($err != 0) or ($snapstatus != 0)) {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot exit in guest. Error 0x%x : exit code 0x%x\n", $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);
		} 
		
		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : exit code 0x%x\n", $gos_guestVssPgmPath, $cmdArgs, $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);

		$vmName =~ s/[^a-zA-Z0-9-.#!@%_]*//g;
		$guestLogFile = $gos_guestPath . "\\VssLogSVR.log";
		$hostLogFile = "/usr/bp/logs.dir/vprotect_" . $vmName . ".vssSvr.log";
		$err = VMCopyFileFromGuestToHost($gos_vmHandle, $guestLogFile, $hostLogFile, 0, VIX_INVALID_HANDLE);
		dbg_log("VMCopyFile of : " . $guestLogFile . " returns " . $err);

		$guestLogFile = $gos_guestPath . "\\VssLogClnt.log";
		$hostLogFile = "/usr/bp/logs.dir/vprotect_" . $vmName . ".vssClnt.log";
		$err = VMCopyFileFromGuestToHost($gos_vmHandle, $guestLogFile, $hostLogFile, 0, VIX_INVALID_HANDLE);
		dbg_log("VMCopyFile of : " . $guestLogFile . " returns " . $err);

		## Remove our working directory and all contents on the guest.
		$err = VMDeleteDirectoryInGuest($gos_vmHandle, $gos_guestPath,0);
		dbg_log("Deletion of working directory " . $gos_guestPath . " status : " . $err);
	}

	VMLogoutFromGuest($gos_vmHandle);
	HostDisconnect($gos_hostHandle);

	return $snapstatus;	
}

# Helper sub-routine to get the URL using which a file should be uploaded
# from the UB-RS appliance to the VM/guest on the ESX host.
#
# Returns the URL on success, else doesn't return anything.
# Caller should handle the failure scenario.
sub get_url_for_file_upload {

	my $vm_view = shift;
	my $guest_authentication = shift;
	my $file_manager = shift;
	my $file_destination_in_guest = shift;
	my $guest_file_attributes = shift;
	my $file_size = shift;
	my $overwrite = shift;

	my $file_upload_url;

	eval {
		$file_upload_url = $file_manager->InitiateFileTransferToGuest(vm =>$vm_view,
						auth => $guest_authentication,
						guestFilePath => $file_destination_in_guest,
						fileAttributes => $guest_file_attributes,
						fileSize => $file_size,
						overwrite => $overwrite);
	};
	if ($@) {
		dbg_log("Error: Failed to create URL for file upload to guest OS.");
		if (ref($@) eq 'SoapFault') {
			dbg_log("Reason: " . $@->fault_string);
		} else {
			dbg_log("Error is: " . Dumper($@));
		}
		return;
	}

	return $file_upload_url;

}

# Helper sub-routine to upload a file to the VM/guest from the UB-RS appliance
# Returns 1 on success, 0 on failure
sub upload_file_to_guest {

	my $file_name = shift;
	my $file_upload_url = shift;

	my $status = 0; # Failed

	my $basename = basename($file_name);
	if (!open(FILE, $file_name)) {
		dbg_log("Error: Failed to open GetOsVer.bat file.");
		return $status;
	}
	my $content = do {
			local $/;
			<FILE> };
	close(FILE);

	my $user_agent = LWP::UserAgent->new();
	my $http_request = HTTP::Request->new(PUT => $file_upload_url);
	$http_request->content($content);
	my $http_result = $user_agent->request($http_request);

	if($http_result->is_success()) {
		dbg_log("Successfully uploaded GetOsVer.bat file to guest.");
		$status = 1; # Success
	} else {
		dbg_log("Upload of file $file_name failed.");
		dbg_log("Error: " . $http_result->error_as_HTML);
	}

	return $status;

}

# Helper sub-routine to get the URL using which a file should be downloaded
# from the VM/guest on the ESX host to the UB-RS appliance.
#
# Returns the URL on success, else doesn't return anything.
# Caller should handle the failure scenario.
sub get_url_for_file_download {

	my $vm_view = shift;
	my $guest_authentication =shift;
	my $file_manager = shift;
	my $guest_log_file = shift;

	my $file_transfer_info;
	my $download_url;

	eval {
		$file_transfer_info = $file_manager->InitiateFileTransferFromGuest(vm =>$vm_view,
								auth => $guest_authentication,
								guestFilePath => $guest_log_file);
	};
	if ($@) {
		dbg_log("Error: Failed to get download information for file download from guest OS.");
		if (ref($@) eq 'SoapFault') {
			dbg_log("Reason: " . $@->fault_string);
		} else {
			dbg_log("Error is: " . Dumper($@));
		}
		return;
	}

	if (not defined $file_transfer_info->url) {
		dbg_log("Error: Failed to get URL for file download.");
		return;
	}

	$download_url = $file_transfer_info->url;

	return $download_url;

}

# Helper sub-routine to download a file from the VM/guest to the UB-RS appliance.
# Returns 1 on success, else doesn't return anything.
sub download_file_from_guest {

	my $guest_log_file = shift;
	my $host_log_file = shift;
	my $download_url = shift;

	my $basename = basename($guest_log_file);
	my $user_agent = LWP::UserAgent->new();
	my $response = $user_agent->get($download_url);
	my $content = $response->decoded_content;

	if (open(my $fh, '>', $host_log_file)) {
		print {$fh} $content;
	} else {
		dbg_log("Error: Failed to write to $host_log_file file.");
		return;
	}

	return 1;
}

# Helper sub-routine to start a program in the guest/VM.
# On success, returns the Process ID of the program that was started.
# On failure, doesn't return anything. Caller should handle this condition.
sub get_program_pid {

	my $vm_view = shift;
	my $guest_authentication = shift;
	my $process_manager = shift;
	my $program_path = shift;
	my $program_args = shift;
	my $start_minimized = shift;

	my $program_start_spec;
	my $program_pid;

	eval {
		$program_start_spec = GuestWindowsProgramSpec->new(programPath => $program_path,
						arguments => $program_args,
						workingDirectory => "",
						startMinimized => $start_minimized);
		$program_pid = $process_manager->StartProgramInGuest(vm => $vm_view,
						auth => $guest_authentication,
						spec => $program_start_spec);
	};
	if ($@) {
		dbg_log("Error: Failed to start program $program_path $program_args in the guest.");
		if (ref($@) eq 'SoapFault') {
			dbg_log("Reason: " . $@->fault_string);
		} else {
			dbg_log("Error is: " . Dumper($@));
		}
		return;
	}

	return $program_pid;

}

# Helper sub-routine to monitor a Process ID and return the exit code of that process.
#
# The "$time" argument indicates for how long (in seconds) to monitor.
# It's upto the caller to decide this time, as some programs can exit quickly and some won't.
# In the case where it's mandatory to get the exit code and we already know that it can
# take a while, caller can pass "$time=0".
sub get_exit_code_from_pid {

	my $vm_view = shift;
	my $guest_authentication = shift;
	my $process_manager = shift;
	my $program_pid = shift;
	my $time = shift; # Time in seconds indicating how long to monitor
			  # a process/program that was started in the guest.
			  # If 0, then, monitor until the program that was
			  # started in the guest exits.

	my $exit_code;
	my $counter = 0;
	my $monitor_until_guest_program_exits = 0;

	if ($time eq 0) {
		dbg_log("Monitoring guest program until it exits.");
		$monitor_until_guest_program_exits = 1;
	}

	while ($counter <= $time) {
		# ListProcessesInGuest() API returns an array of 
		# GuestProcessInfo[] structures. Monitor the array
		# every 2 seconds to see if 'GetOsVer.bat' completed
		# execution and what it returned.
		my $guest_processes = $process_manager->ListProcessesInGuest(vm =>$vm_view,
							auth => $guest_authentication);
		foreach(@$guest_processes) {
			if ($_->pid eq $program_pid) {
				dbg_log("Process with PID $program_pid found.");
				if (defined $_->exitCode) {
					dbg_log("Exit code: " . $_->exitCode);
					$exit_code = $_->exitCode;
				} else {
					dbg_log("Exit code not received yet. Sleeping for 2 seconds...");
					sleep 2;
				}
				last; # Exit foreach
			}
		}
		if (not defined $exit_code) {
			if ($monitor_until_guest_program_exits ne 1) {
				$counter = $counter + 5;
			}
		} else {
			dbg_log("Exit code of process ID $program_pid is $exit_code");
			#$gos_osLevel = $exit_code;
			last; # Exit while
		}
	}

	return $exit_code;

}

# Helper sub-routine to delete the "Unitrends" directory that was created under
# the Windows installation directory.
# Returns 1 on success, 0 on failure.
sub delete_guest_dir {

	my $vm_view = shift;
	my $guest_authentication = shift;
	my $file_manager = shift;
	my $gos_guest_path = shift;
	my $recursive_deletion = shift;

	eval {
		$file_manager->DeleteDirectoryInGuest(vm =>$vm_view,
				auth => $guest_authentication,
				directoryPath => $gos_guest_path,
				recursive => "true");
	};
	if ($@) {
		dbg_log("Error: Failed to delete directory $gos_guest_path in guest.");
		if (ref($@) eq 'SoapFault') {
			dbg_log("Reason: " . $@->fault_string);
		} else {
			dbg_log("Error is: " . Dumper($@));
		}
		return 0;
	}

	return 1;

}

# New sub-routine that uses APIs from the Perl API set and none of the VIX APIs
# to do application-aware processing.
# In any case where an error is encountered, a non-zero value is returned.
sub guest_vss_operation_new {

	my $command = shift;
	my $vmName = shift;
	my $vmxPath = shift;
	my $server = shift;
	my $hostUser = shift;
	my $hostPasswd = shift;
	my $vmUser = shift;
	my $vmPasswd = shift;
	my $backupStatus = shift;
	my $warning_ref = shift;
	my $vm_view = shift;

	my $err = 1; # Return value
	my $cmdArgs;
	my $snapstatus;

	my $winDir; 		# Directory under which Windows installation resides
	my $arch; 		# Architecture of the guest - 32-bit v/s 64-bit
	my $env_variables; 	# Complete set of environment variables in the guest
	my $vProtectVssExeFile; # Path to vProtectVSS EXE file on the appliance

	my $start_minimized = "true"; # Any program that's run in the guest
				      # will be in minimized mode
	my $interactive_session = "false"; # No user interaction needed when validating
					   # credentials for the guest

	my $guest_operations_manager = Vim::get_view(mo_ref => Vim::get_service_content()->guestOperationsManager);
	if (not defined $guest_operations_manager) {
		dbg_log("Error: Unable to find guest_operations_manager object.");
		$$warning_ref .= "\nUnable to retrieve information from the vCenter/ESX server.";
		return $err;
	}

	my $auth_manager = Vim::get_view(mo_ref => $guest_operations_manager->authManager);
	if (not defined $auth_manager) {
		dbg_log("Error: Unable to get authManager object.");
		$$warning_ref .= "\nUnable to retrieve information from the vCenter/ESX server.";
		return $err;
	}

	my $guest_authentication = NamePasswordAuthentication->new(username => $vmUser,
						password => $vmPasswd,
						interactiveSession => $interactive_session);
	eval {
		dbg_log("Validating guest credentials for app-aware processing.");
		$auth_manager->ValidateCredentialsInGuest(vm => $vm_view,
							auth => $guest_authentication);
	};
	if ($@) {
		dbg_log("Error: Failed to login to guest.");
		my $login_err_msg = "\nFailed to login to guest $vmName for VSS operations.";
		if (ref($@) eq 'SoapFault') {
			dbg_log("Reason: " . $@->fault_string);
			dbg_log("FaultType: " . ref($@->detail));
			dbg_log("Dumper -\n" . Dumper($@));

			if (ref($@->detail) eq 'GuestOperationsUnavailable') {
				$$warning_ref .= "$login_err_msg\nVerify that latest version of " .
						"VMware Tools is installed and running.";
			} elsif (ref($@->detail) eq 'InvalidGuestLogin') {
				$$warning_ref .= "$login_err_msg\nVerify that the correct set of " .
						"credentials are used with the appropriate permissions.";
			} else {
				$$warning_ref .= $login_err_msg;
			}

		} else {
			dbg_log("Error is: " . Dumper($@));
			$$warning_ref .= $login_err_msg;
		}
		return $err;
	}

	dbg_log("Successfully logged in to the guest for app-aware processing.");

	my $process_manager = Vim::get_view(mo_ref => $guest_operations_manager->processManager);
	if (not defined $process_manager) {
		dbg_log("Error: Unable to get processManager object.");
		$$warning_ref .= "\nUnable to retrieve information from the vCenter/ESX server.";
		return $err;
	}

	eval {
		$env_variables = $process_manager->ReadEnvironmentVariableInGuest(vm =>$vm_view,
								auth => $guest_authentication);
	};
	if ($@) {
		dbg_log("Error: Failed to get Environment Variable information for the guest.");
		if (ref($@) eq 'SoapFault') {
			dbg_log("Reason: " . $@->fault_string);
		} else {
			dbg_log("Error is: " . Dumper($@));
		}
		$$warning_ref .= "\nFailed to get environment variable information in the guest " . $vmName . ".";
		return $err;
	}

	# Traverse over the list of environment variables to find the variables that we need
	# Environment variables are returned in the form - variable=value
	# Variables that we are looking for:
	# 	windir=<someValue>
	# 	PROCESSOR_ARCHITECTURE=<someValue>
	my $windowsdir_string = "windir";
	my $processor_architecture_string = "PROCESSOR_ARCHITECTURE";
	foreach(@$env_variables) {
		my ($variable, $val) = split /=/, $_;
		if ($variable =~ /^$windowsdir_string/i) {
			dbg_log("Windows directory found - $_");
			$winDir = $val;
			next;
		}
		if ($variable =~ /^$processor_architecture_string/i) {
			dbg_log("Processor architecture found - $_");
			$arch = $val;
			next;
		}
	}

	# No Windows directory? No problem.
	# Warn the backup an continue without performing the VSS snap.
	if ($winDir eq "") {
		dbg_log("Error: Unable to find location of directory where Windows is installed.");
		$$warning_ref = $$warning_ref . "\nFailed to find Windows directory in guest.";
		return $err;
	}

	if ($arch eq "") {
		$arch = "X86";
	}

	my $file_manager = Vim::get_view(mo_ref => $guest_operations_manager->fileManager);
	if (not defined $file_manager) {
		dbg_log("Error: Unable to get fileManager object.");
		$$warning_ref .= "\nUnable to retrieve information from the vCenter/ESX server.";
		return $err;
	}

	## Place the VSS agent in %WinDir%\Windows\Unitrends
	$gos_guestPath = $winDir . "\\Unitrends";

	if ($command eq "create") {
		# Create Unitrends directory
		my $directory_creation_successful = 0;
		my $createParentDirectories = "true";
		eval {
			$file_manager->MakeDirectoryInGuest(vm => $vm_view,
					auth => $guest_authentication,
					directoryPath => $gos_guestPath,
					createParentDirectories => $createParentDirectories);
		};
		if ($@) {
			dbg_log("Error: Failed to create Unitrends directory in guest.");
			if (ref($@) eq 'SoapFault') {
				dbg_log("Reason: " . $@->fault_string);
				if (ref($@->detail) eq 'FileAlreadyExists') {
					# Don't error out if the Unitrends directory already exists
					dbg_log("Info: Guest directory $gos_guestPath already exists.");
					$directory_creation_successful = 1;
				}
			} else {
				dbg_log("Error is: " . Dumper($@));
			}
			if ($directory_creation_successful eq 0) {
				$$warning_ref .= "\nFailed to create Unitrends directory in the guest.";
				return $err;
			}
		}
		dbg_log("Successfully created Unitrends directory in guest at $gos_guestPath.");
	}

	# Now copy over the 'GetOsVer.bat' file to the guest
	my $getosversion_file = "/usr/bp/bin/vprotect/GetOsVer.bat";
	my $getosversion_file_size = -s $getosversion_file;
	my $file_destination_in_guest = $gos_guestPath . "\\GetOsVer.bat";

	my $overwrite = "true";
	my $guest_file_attributes = GuestFileAttributes->new();

	my $file_upload_url;
	my $file_upload_status;

	$file_upload_url = get_url_for_file_upload($vm_view, $guest_authentication, $file_manager,
					$file_destination_in_guest, $guest_file_attributes, $getosversion_file_size,
					$overwrite);
	if (not defined $file_upload_url) {
		dbg_log("Error: Failed to get URL for GetOsVer.bat file upload.");
		$$warning_ref .= "\nFailed to copy files to guest for VSS operations.";
		return $err;
	}

	$file_upload_url =~ s/\*/$server/;
	dbg_log("URL to copy GetOsVer.bat file to guest: $file_upload_url");

	$file_upload_status = upload_file_to_guest($getosversion_file, $file_upload_url);
	if ($file_upload_status == 1) {
		dbg_log("File $getosversion_file uploaded successfully to guest.");
	} else {
		dbg_log("Error: Failed to upload file $getosversion_file.");
		$$warning_ref .= "\nFailed to copy files to guest for VSS operations.";
		return $err;
	}

	# GetOsVer.bat file uploaded successfully.
	# Now run the same to get the OS version.
	my $progPath = $winDir . "\\system32\\cmd.exe ";
	my $batchProgArgs = "/C " . $gos_guestPath . "\\GetOsVer.bat";

	my $program_pid = get_program_pid($vm_view, $guest_authentication, $process_manager,
				$progPath, $batchProgArgs, $start_minimized);
	if (not defined $program_pid) {
		dbg_log("Error running $progPath $batchProgArgs in guest.");
		$$warning_ref .= "\nFailed to run program in guest for VSS operations.";
		return $err;
	}

	my $time;
	my $exit_code;
	$gos_osLevel = 6;

	# Need to do the below, since the guestProcess structure doesn't get populated
	# with the exitCode of 'GetOsVer.bat' immediately after the program is run
	$time = 15; # Total time in seconds for how long to monitor
	$exit_code = get_exit_code_from_pid($vm_view, $guest_authentication, $process_manager,
				$program_pid, $time);
	$gos_osLevel = $exit_code;

	# If Windows 2003 or earlier always run the 32-bit binaries.
	if ($gos_osLevel == 5) {
		$arch = "X86";
	}

	# Copy the VSS program over.
	if ($arch eq "AMD64") {
		$vProtectVssExeFile = "/usr/bp/bin/vprotect/VProtectVss64.exe";
		$gos_guestVssPgmPath = $gos_guestPath . "\\VProtectVss64.exe";
	} else {
		$vProtectVssExeFile = "/usr/bp/bin/vprotect/VProtectVss.exe";
		$gos_guestVssPgmPath = $gos_guestPath . "\\VProtectVss.exe";
	}

	if ($command eq "create") {
		# Creating the snapshot

		# Now, upload the $vProtectVssExeFile from the UB-RS appliance to the guest/VM.
		$file_upload_url = undef;
		$guest_file_attributes = undef;
	
		my $vProtectVssExeFileSize = -s $vProtectVssExeFile;
		$guest_file_attributes = GuestFileAttributes->new();

		$file_upload_url = get_url_for_file_upload($vm_view, $guest_authentication, $file_manager,
					$gos_guestVssPgmPath, $guest_file_attributes, $vProtectVssExeFileSize,
					$overwrite);
		if (not defined $file_upload_url) {
			dbg_log("Error: Failed to get URL for $vProtectVssExeFile file upload.");
			$$warning_ref .= "\nFailed to copy files to guest for VSS operations.";
			return $err;
		}

		$file_upload_url =~ s/\*/$server/;
		dbg_log("URL for $vProtectVssExeFile upload to guest: $file_upload_url");

		$file_upload_status = 0;
		$file_upload_status = upload_file_to_guest($vProtectVssExeFile, $file_upload_url);
		if ($file_upload_status == 1) {
			dbg_log("File $vProtectVssExeFile uploaded successfully to guest.");
		} else {
			dbg_log("Error: Failed to upload file $vProtectVssExeFile.");
			$$warning_ref .= "\nFailed to copy files to guest for VSS operations.";
			return $err;
		}

		dbg_log("$vProtectVssExeFile program copied successfully to guest.");

		# Run the VSS program in server mode first. Leaving it running awaiting commands
		# to create vss snapshots and then later remove them and exit.
		$cmdArgs = " Server  " . $gos_guestPath . "\\VssLogSVR.log";

		$program_pid = undef;
		$program_pid = get_program_pid($vm_view, $guest_authentication, $process_manager,
					$gos_guestVssPgmPath, $cmdArgs, $start_minimized);
		if (not defined $program_pid) {
			dbg_log("Error running $gos_guestVssPgmPath $cmdArgs in guest.");
			$$warning_ref .= "\nFailed to run program in guest for VSS operations.";
			return $err;
		}

		dbg_log("Successfully executed $gos_guestVssPgmPath $cmdArgs.");
		dbg_log("PID in guest: $program_pid");

		dbg_log("Creating VSS snapshot.");

		$cmdArgs = " CreateSnap " . $gos_guestPath . "\\VssLogClnt.log";
		$program_pid = undef;
		$program_pid = get_program_pid($vm_view, $guest_authentication, $process_manager,
					$gos_guestVssPgmPath, $cmdArgs, $start_minimized);
		if (not defined $program_pid) {
			dbg_log("Error running $gos_guestVssPgmPath $cmdArgs in guest.");
			$$warning_ref .= "\nFailed to run program in guest for VSS operations.";
			return $err;
		}

		dbg_log("Successfully executed $gos_guestVssPgmPath $cmdArgs.");
		dbg_log("PID in guest: $program_pid");

		$snapstatus = undef;

		# Depending on the activity in the guest OS, it can take
		# some time to complete the create snapshot operation.
		$time = 0; # Monitor until the process completes the Create Snapshot operation

		$snapstatus = get_exit_code_from_pid($vm_view, $guest_authentication, $process_manager,
					$program_pid, $time);

		dbg_log("Exit code for Vss program for createSnap: $snapstatus");
		
		if ($snapstatus != 0)  {
			dbg_log("VSS snapshot creation failed.");
			## Display Windows hex error code.
			if ($snapstatus == 1) {
				$$warning_ref .= "\nFailed to run vss snapshot create program in guest. Guest VSS snapshots are not supported for guest OS.\n";
			} else {	
				$$warning_ref .= sprintf("\n Failed to run vss snapshot create program in guest. Error 0x%x, VSS program returns error 0x%x\n", $err & 0xFFFFFFFF,  $snapstatus & 0xFFFFFFFF);
			}
		} else {
			dbg_log("VSS snapshot created successfully...");
		}

	} else {
		## Removing the VSS snapshot
		my $hostLogFile;
		my $guestLogFile;
		my $time;
		$snapstatus = 0;

		if ($backupStatus eq "SUCCESS") {
			$cmdArgs = " ReleaseOk " . $gos_guestPath . "\\VssLogClnt.log";
		} else {
			$cmdArgs = " ReleaseFail " . $gos_guestPath . "\\VssLogClnt.log";
		}

		dbg_log("gos_guestVssPgmPath: $gos_guestVssPgmPath, cmdArgs: $cmdArgs");

		$program_pid = undef;
		$program_pid = get_program_pid($vm_view, $guest_authentication, $process_manager,
					$gos_guestVssPgmPath, $cmdArgs, $start_minimized);
		if (not defined $program_pid) {
			dbg_log("Error running $gos_guestVssPgmPath $cmdArgs in guest.");
			$$warning_ref .= "\nFailed to run program in guest for removing VSS snapshot.";
			return $err;
		}

		dbg_log("PID for VssVprotect EXE for Remove Snapshot: $program_pid");

		# Depending on the activity in the guest OS, it can take
		# some time to complete the remove snapshot operation.
		$time = 0; # Monitor until the process completes the Release Snapshot operation

		$snapstatus = get_exit_code_from_pid($vm_view, $guest_authentication, $process_manager,
					$program_pid, $time);

		dbg_log("Exit code for Vss program for Release snapshot: $snapstatus");

		if ($snapstatus != 0) {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot release in guest. Error 0x%x : exit code 0x%x\n", $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);
		} 
		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : exit code 0x%x\n", $gos_guestVssPgmPath, $cmdArgs, $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);

		$cmdArgs = " Exit " . $gos_guestPath . "\\VssLogClnt.log";
		dbg_log("gos_guestVssPgmPath: $gos_guestVssPgmPath, cmdArgs: $cmdArgs");

		$program_pid = undef;
		$program_pid = get_program_pid($vm_view, $guest_authentication, $process_manager,
					$gos_guestVssPgmPath, $cmdArgs, $start_minimized);
		if (not defined $program_pid) {
			dbg_log("Error running $gos_guestVssPgmPath $cmdArgs in guest.");
			$$warning_ref .= "\nFailed to run program in guest for removing VSS snapshot.";
			return $err;
		}

		dbg_log("PID for VssVprotect EXE for Exit: $program_pid");

		$snapstatus = get_exit_code_from_pid($vm_view, $guest_authentication, $process_manager,
					$program_pid, $time);

		dbg_log("Exit code for Vss program for Exit: $snapstatus");

		if ($snapstatus != 0) {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot exit in guest. Error 0x%x : exit code 0x%x\n", $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);
		} 
		
		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : exit code 0x%x\n", $gos_guestVssPgmPath, $cmdArgs, $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);

		my $download_status;
		$vmName =~ s/[^a-zA-Z0-9-.#!@%_]*//g;
		$guestLogFile = $gos_guestPath . "\\VssLogSVR.log";
		$hostLogFile = "/usr/bp/logs.dir/vprotect_" . $vmName . ".vssSvr.log";

		my $file_download_url = get_url_for_file_download($vm_view, $guest_authentication,
						$file_manager, $guestLogFile);
		if (not defined $file_download_url) {
			# Just log. Don't fail/warn, since the Backup is successful
			dbg_log("Error: Failed to get URL for $guestLogFile file download.");
		} else {
			#$file_download_url =~ s/\*/$server/;
			dbg_log("Download URL is: $file_download_url");
			$download_status = download_file_from_guest($guestLogFile, $hostLogFile, $file_download_url);
			if (not defined $download_status) {
				# Just log. Don't fail/warn, since the Backup is successful
				dbg_log("Warning: Failed to download $guestLogFile.");
			}
		}

		$guestLogFile = $gos_guestPath . "\\VssLogClnt.log";
		$hostLogFile = "/usr/bp/logs.dir/vprotect_" . $vmName . ".vssClnt.log";

		$file_download_url = undef;
		$file_download_url = get_url_for_file_download($vm_view, $guest_authentication,
						$file_manager, $guestLogFile);
		if (not defined $file_download_url) {
			# Just log. Don't fail/warn, since the Backup is successful
			dbg_log("Warning: Failed to get URL for $guestLogFile file download.");
		} else {
			#$file_download_url =~ s/\*/$server/;
			dbg_log("Download URL for $guestLogFile: $file_download_url");
			$download_status = undef;
			$download_status = download_file_from_guest($guestLogFile, $hostLogFile, $file_download_url);
			if (not defined $download_status) {
				# Just log. Don't fail/warn, since the Backup is successful
				dbg_log("Warning: Failed to download $guestLogFile.");
			}
		}

		# Now delete the directory that we had created in the guest
		my $recursive_deletion = "true";
		my $deletion_status = delete_guest_dir($vm_view, $guest_authentication, $file_manager,
						$gos_guestPath, $recursive_deletion);
		if ($deletion_status eq 0) {
			dbg_log("Warning: Failed to delete Unitrends directory.");
		} else {
			dbg_log("Deletion of Unitrends directory was successful.");
		}

	}

	return $snapstatus;	

}

# kas 9/16.2012
# Added.
sub do_guest_vss_snapshot {
	my $vmuuid = shift;
	my $server = shift;
	my $hostUser = shift;
	my $hostPasswd = shift;
	my $vmUser = shift;
	my $vmPasswd = shift;
	my $backupStatus = shift;
	my $operation = shift;
	my $warning_ref = shift;
	my $esx_version_on_which_vm_exists = shift;

	my $vm;
	my $guestObj;
	my $vmxPath;
	my $err;
	my $pwrState;
	my $vmName;
	my $warning = "";

	dbg_log("Entering create_guest_vss_snapshot");

	my $vmmoref = VProtect::vim_get_vm_by_uuid($vmuuid);
	$vm = Vim::get_view(mo_ref => $vmmoref);

	if (not defined ($vm->config)) {
		$$warning_ref = $$warning_ref . "\n Unable to retrieve VM configuration. (Config not defined.)\n";
		return 1;
	}

	$vmName = $vm->name;

	$pwrState = $vm->runtime->powerState->val;
	# Fetch path to vmx file for later use by guest vss operations
	$vmxPath = $vm->config->files->vmPathName;
	dbg_log("VM " . $vmName . " : power state is " . $pwrState);

	my ($major, $minor, $micro) = split /\./, $esx_version_on_which_vm_exists;
	dbg_log("ESX version on which VM exists: $esx_version_on_which_vm_exists");
	dbg_log("ESX version major: $major, minor: $minor, micro: $micro");

	if ($pwrState eq 'poweredOn') {
		dbg_log("VM " . $vmName . " powered on: path: " . $vmxPath);
		if ($major >= 6) {
			if ( ($minor >= 5) or ($major > 6) ) {
				dbg_log("Performing VSS operations using non-VIX APIs.");
				$err = guest_vss_operation_new($operation, $vmName, $vmxPath, $server, $hostUser, $hostPasswd, $vmUser, $vmPasswd, $backupStatus, \$warning, $vm);
			} else {
				dbg_log("Performing VSS operations using VIX APIs for ESX 6.0.0.");
				$err = guest_vss_operation($operation, $vmName, $vmxPath, $server, $hostUser, $hostPasswd, $vmUser, $vmPasswd, $backupStatus, \$warning);
			}
		} else {
			dbg_log("Performing VSS operations using VIX APIs.");
			$err = guest_vss_operation($operation, $vmName, $vmxPath, $server, $hostUser, $hostPasswd, $vmUser, $vmPasswd, $backupStatus, \$warning);
		}
		if ($err != 0) {
			$$warning_ref .= $warning;
		}
	} else {
		printf STDERR "VM " . $vmName . " not powered - skipping guest VSS snapshot.\n";
		$$warning_ref = $$warning_ref . "\nVirtual machine is powered OFF. Unable to perform application-aware processing.";
		$err = 1;
	}

	return $err;
}

sub do_backup {
	my $backuptype = shift;
	my $vmuuid = shift;
	my $change_ids = shift;
	my $prev_hw_hash = shift;
	my $serverAddr = shift;
	my $appAwareFlag = shift;
	my $hostUserName = shift;
	my $hostPassword = shift;
	my $gosUserName = shift;
	my $gosPassword = shift;
	my $keys = shift;	# Exclude Keys
	my $is_template = shift;
	my $quiesce = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $warning = "";
	my $info = "";

	my $vm_moref;
	my $snap_moref;
	my $err = 0;
	my $retry = 0;

	dbg_log("Original keys of disks to exclude - " . $keys);

	if ((defined ($change_ids)) and ($is_template eq "YES")) {
		die "Incremental and Differential Backups are not supported for VMware Templates.\n";
	}

	my $promote_next_backup_because_cbt_failed = 0;
	my $promote_next_backup_because_cbt_off = 0; # Used for Incr/Diff Backups

	my $change_id_hash = split_change_ids($change_ids);

	my $key_array = split_keys($keys);

	if (not defined($key_array)) {
		$key_array = "";
	}

	dbg_log("Got args:  appAware: " . $appAwareFlag . " : gosUser : " . $gosUserName);
	dbg_log("Backup entity is a Template - $is_template");
	dbg_log("Snapshotting with quiesce - $quiesce");

	if ($is_template eq "NO") {
		# Get a moref for the VM.
		$vm_moref = get_moref_by_uuid($vmuuid, $xmldoc, $xmlroot);
	} else {
		# Get a moref for the Template
		$vm_moref = get_template_moref_by_uuid($vmuuid, $xmldoc, $xmlroot);
	}

	my ($new_hw_hash, $hw_hash_using_only_disks) = VProtect::get_vm_information_by_moref($vm_moref, $key_array, $xmldoc, $xmlroot);

	dbg_log("H/W hash calculated using all the virtual hardware configuration: $new_hw_hash");
	dbg_log("H/W hash calculated using only virtual disks: $hw_hash_using_only_disks");
	if (defined $prev_hw_hash) {
		dbg_log("H/W hash passed from the database from the previous Backup: $prev_hw_hash");
	}

	my $is_full = 1;
	if (defined $change_ids) {
		$is_full = 0;
	}

	validate_hw_hash($prev_hw_hash, $new_hw_hash, $is_full, $xmldoc, $xmlroot, $vm_moref, $hw_hash_using_only_disks);

	check_for_fault_tolerance($vm_moref);

	RETRY:
	if ($is_template eq "NO") {
		clean_up_snapshots($vm_moref, $xmldoc, $xmlroot, \$warning);

		configure_change_tracking($vm_moref, $is_full, $xmlroot, \$warning, $retry, \$promote_next_backup_because_cbt_off);
	}

	if ($promote_next_backup_because_cbt_off) {
		VProtect::promote_next_reason($xmldoc, 'CBT found OFF for an Incremental/Differential Backup.');
		die "Change Bloack Tracking (CBT) must be enabled before running a Differential or an Incremental backup.\n" .
		"The next Backup of this Virtual Machine will be promoted to a Full to enable CBT.\n";
	}

	# Find the ESX version on which the VM 'currently' resides.
	# This will later help to understand whether to use client side compression
	# or not, since that feature is available only on ESXi 6.5.0 and later.

	# Not adjusting this for 6.7.0 as it seems to deal with compression decisions and its ok if compression is off on newer releases.
	my $esx_version_on_which_vm_exists = VProtect::find_my_esx_version($vmuuid);
	if (not defined $esx_version_on_which_vm_exists) {
		# Don't fail the Backup if ESX version is not found.
		# Assign a value that will work on ESXi versions older than 6.5.0.
		dbg_log("Unable to retrieve ESX server version.");
		$esx_version_on_which_vm_exists = "6.0.0";
	}

	## KAS
	## If guest vss snapshot fails, then update the $warning var to
	## make sure the backup completes minimally with a warning status.
	##
	if ($appAwareFlag eq "YES") {
		if ($is_template eq "YES") {
			die "Cannot do app-aware processing for a Template.\n" .
			"Convert this Template to a Virtual Machine to use app-aware capability.\n";
		} else {
			dbg_log("Doing app-aware guest processing.");
			$err = do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, 0, "create", \$warning, $esx_version_on_which_vm_exists);
			dbg_log("Value of err is $err");
			if ($err != 0) {
				# From 9.1.0, if app-aware processing fails, continue with the Backup.
				# However, the Backup will be completed with a warning
				dbg_log("App-aware guest processing failed.");
				my $remWarn;
				do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, 0, "remove", \$remWarn, $esx_version_on_which_vm_exists);
			}
			dbg_log("Guest vss snapshot warnings:  " . $warning);
		}
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snapshots_exist = snapshots_exist($vm_view);

	my $has_independent_disks = independent_disks_present($vm_view);

	if ($is_template eq "NO") {
		$snap_moref = create_snapshot($vm_moref, $xmldoc, $xmlroot, \$warning, $err, $has_independent_disks, $quiesce, $appAwareFlag, \$info);
	}

	if (!(defined $snap_moref) and ($is_template eq "NO")) {
		dbg_log("create_snapshot returns empty moref");
	}

	# Use the default/best transport available transport mode or NBD
	transport_mode($xmlroot);

	dbg_log("ESX version on which the VM exists currently - $esx_version_on_which_vm_exists");
	$xmlroot->appendTextChild('VMESXVersion', $esx_version_on_which_vm_exists);	

	my @diskinfo;
	eval {
		@diskinfo = get_disk_info($vm_moref, $snap_moref, $change_id_hash, $key_array, $is_template, $xmldoc, $xmlroot, $is_full, \$promote_next_backup_because_cbt_failed, $snapshots_exist);

		dbg_log("Return value from get_disk_info is -\n" .Dumper(@diskinfo));
		if ($diskinfo[0] == 33) {
			if ($retry == 0) {
				dbg_log("Removing Unitrends_backup snapshot to retry");
				$retry = $retry + 1;
				VProtect::remove_snapshot($snap_moref->{'value'}, $vmuuid);
				goto RETRY;
			} else {
				dbg_log("Not making another attempt to get CBT information.");
			}
		}

	};
	if ($@) {
		my @old_error = $@;
		if ($is_template eq "NO") {
			$err = do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, 0, "remove", \$warning, $esx_version_on_which_vm_exists);
			# Remove snapshot and throw again
			dbg_log("get_disk_info() failed, removing snapshot with moref " . $snap_moref->{'value'});
			VProtect::remove_snapshot($snap_moref->{'value'}, $vmuuid);
			if ($promote_next_backup_because_cbt_failed) {
				VProtect::promote_next_reason($xmldoc, 'A VM changed blocked tracking error has occurred.');
			}
		}
		die (@old_error);
	}

	if ($is_template eq "NO") {
		my $snap_xmlelem = $xmldoc->createElement('snapshot');
		$snap_xmlelem->appendTextChild('snap_moref', $snap_moref->{'value'});
		$snap_xmlelem->appendTextChild('snap_moref_type', $snap_moref->{'type'});
		$xmlroot->addChild($snap_xmlelem);
	}

	if ($warning ne "") {
		# Put any warnings we generated into the XML.
		VProtect::output_xml_warning($xmldoc, $warning);
	}

	if ($info ne "") {
		# Put any info messages we generated also in the XML.
		VProtect::output_xml_info($xmldoc, $info);
	} else {
		dbg_log("No information messages for the XML being returned.");
	}

}

sub do_backup_conclude {
	my $vmuuid = shift;
	my $serverAddr = shift;
	my $appAwareFlag = shift;
	my $hostUserName = shift;
	my $hostPassword = shift;
	my $gosUserName = shift;
	my $gosPassword = shift;
	my $backupStatus = shift;
	my $warning = "";
	my $vm_moref;
	my $snap_moref;
	my $err = 0;

	dbg_log("Got args:  appAware: " . $appAwareFlag . " : gosUser : " . $gosUserName);

	my $esx_version_on_which_vm_exists = VProtect::find_my_esx_version($vmuuid);
	if (not defined $esx_version_on_which_vm_exists) {
		# Don't fail the Backup if ESX version is not found.
		# Assign a value that will work on ESXi versions older than 6.5.0.
		dbg_log("Unable to retrieve ESX server version.");
		$esx_version_on_which_vm_exists = "6.0.0";
	}

	if ($appAwareFlag eq "YES") {
		# Get a moref for the VM.
		dbg_log("About to call into do_guest_vss_snapshot");
		$err = do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, $backupStatus, "release", \$warning, $esx_version_on_which_vm_exists);
	}
	
	dbg_log("Leaving do_backup_conclude");

	return $err;
}

1;
